package it.arnaldo.unibs.tamagolem;

import java.util.*;

public class WorldBuilder {

    public static ElementGraph buildWorld() {
        // genero un numero casuale che mi dice quanti elementi prendo
        Element[] allElements = Element.values();
        int numElements = (int)(Math.random() * 8) + 3;

        List<Element> selectedElements = new ArrayList<>(Arrays.asList(allElements));
        Collections.shuffle(selectedElements);

        // seleziono la sublist di lunghezza numElements
        selectedElements = selectedElements.subList(0, numElements);


        // Mappa delle direzioni tra elementi (1 = attacca, -1 = subisce)
        Map<Element, Map<Element, Integer>> linkDirections = generateValidDirections(selectedElements);



        // Mappa indice per ogni coppia (non ordinata) di elementi
        Map<Set<Element>, Integer> pairIndex = new HashMap<>();
        int index = 0;
        for (int i = 0; i < numElements; i++) {
            for (int j = i + 1; j < numElements; j++) {
                pairIndex.put(Set.of(selectedElements.get(i), selectedElements.get(j)), index++);
            }
        }

        // Costruzione del sistema lineare Ax = b (b = 0)
        int[][] A = new int[numElements][pairIndex.size()];
        int[] b = new int[numElements];

        for (int i = 0; i < numElements; i++) {
            Element e1 = selectedElements.get(i);
            int[] row = new int[pairIndex.size()];
            for (Element e2 : selectedElements) {
                if (e1 == e2) continue;
                int dir = linkDirections.get(e1).get(e2);
                int pos = pairIndex.get(Set.of(e1, e2));
                row[pos] = dir;
            }
            A[i] = row;
            b[i] = 0;
        }

        // sistema lineare per trovare le soluzioni che verificano l'equilibrio

        try {
            LinearSystem system = new LinearSystem(numElements, A, b);
            int[] solution = system.resolve();

            // dalle soluzioni del sistema popolo il grafo
            ElementGraph world = new ElementGraph();
            for (Element e1 : selectedElements) {
                GraphElement graphElement = new GraphElement(e1);
                for (Element e2 : selectedElements) {
                    if (e1 == e2) continue;
                    int dir = linkDirections.get(e1).get(e2);
                    int val = solution[pairIndex.get(Set.of(e1, e2))];
                    graphElement.addLink(new ElementLink(dir * val, e2));
                }
                world.addElement(graphElement);
            }

            world.printGraph();
            return world;

         }catch (Exception e){
            buildWorld();
        }


        return null;

    }

    private static Map<Element, Map<Element, Integer>> generateValidDirections(List<Element> elements) {
        int n = elements.size();
        Random rand = new Random();
        Map<Element, Map<Element, Integer>> linkDirections = new HashMap<>();

        for (Element e : elements) {
            linkDirections.put(e, new HashMap<>());
        }

        // creo le varie coppie (l'ordine conta successivamente)
        List<Pair> pairs = new ArrayList<>();
        for (int i = 0; i < n; i++) {
            for (int j = i + 1; j < n; j++) {
                pairs.add(new Pair(elements.get(i), elements.get(j)));
            }
        }

        boolean valid;
        // controllo che ci sia almeno un arco entrante e uno uscente
        do {
            valid = true;
            Map<Element, Integer> exitingArcs = new HashMap<>();
            Map<Element, Integer> enteringArcs = new HashMap<>();

            // resetto i valori per evitare che cicli precedenti abbiano variato qualcosa
            for (Element e : elements) {
                linkDirections.get(e).clear();
                exitingArcs.put(e, 0);
                enteringArcs.put(e, 0);
            }

            // mescolo le coppie e determino i collegamenti (1 uscente, -1 entrante)
            Collections.shuffle(pairs);
            for (Pair pair : pairs) {
                Element e1 = pair.getFirst();
                Element e2 = pair.getSecond();
                boolean dir = rand.nextBoolean();
                if (dir) {
                    linkDirections.get(e1).put(e2, 1);
                    linkDirections.get(e2).put(e1, -1);
                    enteringArcs.put(e1, enteringArcs.get(e1) + 1);
                    exitingArcs.put(e2, exitingArcs.get(e2) + 1);
                } else {
                    linkDirections.get(e2).put(e1, 1);
                    linkDirections.get(e1).put(e2, -1);
                    enteringArcs.put(e2, enteringArcs.get(e2) + 1);
                    exitingArcs.put(e1, exitingArcs.get(e1) + 1);
                }
            }

            // controllo se la struttura è valida
            for (Element e : elements) {
                if (exitingArcs.get(e) == 0 || enteringArcs.get(e) == 0) {
                    valid = false;
                    break;
                }
            }

        } while (!valid);

        return linkDirections;
    }
}
